--------------------------------------------------------------------------------
-- COUNT Iterations of algorithm!
-- VHDL Test Bench for module: fermat_divider.vhd (ch4) 
--
-- Executes NUMBER_TESTS operations with random values. 
--
--------------------------------------------------------------------------------
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE IEEE.std_logic_arith.all;
USE ieee.std_logic_unsigned.all;
USE ieee.numeric_std.ALL;
USE ieee.std_logic_textio.ALL;
use ieee.math_real.all; -- for UNIFORM, TRUNC

USE std.textio.ALL;
use work.Fermat_divider_parameters.all;

ENTITY count_iterations_fermat_divider IS
END count_iterations_fermat_divider;

ARCHITECTURE behavior OF count_iterations_fermat_divider IS 

  -- Component Declaration for the Unit Under Test (UUT)
  COMPONENT Fermat_divider is
  PORT (
      x, y: in std_logic_vector(K-1 downto 0);
      clk, reset, start: in std_logic;  
      z: out std_logic_vector(K-1 downto 0);
      done: out std_logic
      );
  END COMPONENT;

  -- Internal signals
  SIGNAL x, y, z :  std_logic_vector(K-1 downto 0) := (others=>'0');
  SIGNAL clk, reset, start, done: std_logic;

  constant DELAY : time := 100 ns;
  constant PERIOD : time := 200 ns;
  constant DUTY_CYCLE : real := 0.5;
  constant OFFSET : time := 0 ns;
  constant NUMBER_TESTS: natural := 10;

BEGIN

  -- Instantiate the Unit Under Test (UUT)
  uut: Fermat_divider PORT MAP(x => x, y => y,
                          clk => clk, reset => reset, start => start,
                          z => z, done => done);

  PROCESS -- clock process for clk
  BEGIN
  WAIT for OFFSET;
  CLOCK_LOOP : LOOP
     clk <= '0';
     WAIT FOR (PERIOD *(1.0 - DUTY_CYCLE));
     clk <= '1';
     WAIT FOR (PERIOD * DUTY_CYCLE);
  END LOOP CLOCK_LOOP;
  END PROCESS;

  tb_proc : PROCESS --generate values
    PROCEDURE gen_random(X : out std_logic_vector (K-1 DownTo 0); K: natural; s1, s2: inout Natural) IS
      VARIABLE i_x, aux: integer;
      VARIABLE rand: real;
    BEGIN
        aux := K/16;
        for i in 1 to aux loop
          UNIFORM(s1, s2, rand);
          i_x := INTEGER(TRUNC(rand * real(2**16)));
          x(i*16-1 downto (i-1)*16) := CONV_STD_LOGIC_VECTOR (i_x, 16);
        end loop;
        UNIFORM(s1, s2, rand);
        i_x := INTEGER(TRUNC(rand * real(2**(K-aux*16))));
        x(K-1 downto aux*16) := CONV_STD_LOGIC_VECTOR (i_x, (K-aux*16));
    END PROCEDURE;
    VARIABLE TX_LOC : LINE;
    VARIABLE TX_STR : String(1 to 4096);
    VARIABLE seed1, seed2: positive; 
    VARIABLE i_x, i_y, i_p, i_z, i_yz_modp: integer;
    VARIABLE cycles, max_cycles, min_cycles, total_cycles: integer := 0;
    VARIABLE avg_cycles: real;
    VARIABLE initial_time, final_time: time;
    VARIABLE xx: std_logic_vector (K-1 DownTo 0) ;

  BEGIN
    min_cycles:= 2**20;
    start <= '0'; reset <= '1';
    WAIT FOR PERIOD;
    reset <= '0';
    WAIT FOR PERIOD;
    
    for I in 1 to NUMBER_TESTS loop
      gen_random(xx, K, seed1, seed2);
      while xx >= P loop gen_random(xx, K, seed1, seed2);end loop;
      x <= xx;
      gen_random(xx, K, seed1, seed2);
      while xx >= P or xx = ZERO loop gen_random(xx, K, seed1, seed2);end loop;
      y <= xx;

      start <= '1'; initial_time := now;
      WAIT FOR PERIOD;
      start <= '0';
      wait until done = '1';
      final_time := now;
      cycles := (final_time - initial_time)/PERIOD;
      total_cycles := total_cycles+cycles;
      ASSERT (FALSE) REPORT "Number of Cycles: " & integer'image(cycles) & "  TotalCycles: " & integer'image(total_cycles) SEVERITY WARNING;
      if cycles > max_cycles then  max_cycles:= cycles; end if;
      if cycles < min_cycles then  min_cycles:= cycles; end if;


      WAIT FOR 2*PERIOD;
--    NOT TESTING THE ALGORITHM. 
--      i_z := conv_integer(z);
--      i_yz_modp := (i_y * i_z) mod i_p; 
--      IF ( i_x /= i_yz_modp ) THEN 
--        write(TX_LOC,string'("ERROR!!! (X=")); write(TX_LOC, x);
--        ASSERT (FALSE) REPORT TX_STR SEVERITY ERROR;
--      END IF;  
    end loop;
    WAIT FOR DELAY;
    avg_cycles := real(total_cycles)/real(NUMBER_TESTS);
    ASSERT (FALSE) REPORT
    "Simulation successful!.  MinCycles: " & integer'image(min_cycles) &
    "  MaxCycles: " & integer'image(max_cycles) & "  TotalCycles: " & integer'image(total_cycles) &
    "  AvgCycles: " & real'image(avg_cycles)
    SEVERITY FAILURE;
  END PROCESS;

END;
