----------------------------------------------------------------------------------
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use work.mult_subt_parameters.all;

entity mult_subt_container is
    Port ( 
        c, d: in std_logic_vector (31 downto 0);
        q: in std_logic_vector(31 downto 0);
        sel: in std_logic_vector(K/32 downto 0);
        clk, reset, start: in std_logic; 
        z: out std_logic_vector (31 downto 0);
        done: out std_logic);
  
end mult_subt_container;

architecture Behavioral of mult_subt_container is
  COMPONENT mult_subt
  PORT(
  c, d: in std_logic_vector (2*K-1 downto 0);
  q: in std_logic_vector(K-1 downto 0);
  clk, reset, start: in std_logic; 
  z: out std_logic_vector (2*K-1 downto 0);
  done: out std_logic
    );
  END COMPONENT;

signal c_long, d_long, z_long: std_logic_vector (2*K-1 downto 0);
signal q_long: std_logic_vector (K-1 downto 0);

begin


    forgen: for i in 0 to K/32-1 generate
      regs_IN: process(clk)
      begin    
      if clk'event and clk = '1' then
        if sel(i) = '1' then 
          c_long(32*(i+1)-1 downto 32*i) <= c;
          d_long(32*(i+1)-1 downto 32*i) <= d;
          q_long(32*(i+1)-1 downto 32*i) <= q;
        end if;
      end if;
      end process regs_IN;
    end generate;
  Inst_mult_subt: mult_subt PORT MAP(
    c => c_long,
    d => d_long,
    q => q_long(K-1 downto 0),
    clk => clk,
    reset => reset,
    start => start,
    z => z_long,
    done => done
  );

  mux_out: process(z_long, sel)
  begin
    for i in 0 to K/32-1 loop
      if sel(i) = '1' then 
        z <= z_long(32*(i+1)-1 downto 32*i);
      end if;
    end loop;
  end process mux_out;
end Behavioral;

