--------------------------------------------------------------------------------
-- 
-- VHDL Test Bench for module: mult_subt.vhd (ch4) 
--
-- Executes 1000 operations with random values. 
--
--------------------------------------------------------------------------------
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE IEEE.std_logic_arith.all;
USE ieee.std_logic_signed.all;
USE ieee.numeric_std.ALL;
USE ieee.std_logic_textio.ALL;
use ieee.math_real.all; -- for UNIFORM, TRUNC

USE std.textio.ALL;
use work.mult_subt_parameters.all;

ENTITY test_mult_subt IS
END test_mult_subt;

ARCHITECTURE behavior OF test_mult_subt IS 

  -- Component Declaration for the Unit Under Test (UUT)
  COMPONENT mult_subt is
  PORT (
    c, d: in std_logic_vector (2*K-1 downto 0);
    q: in std_logic_vector(K-1 downto 0);
    clk, reset, start: in std_logic; 
    z: out std_logic_vector(2*K-1 downto 0);
    done: inout std_logic
    );
  END COMPONENT;

  --Inputs
  SIGNAL c, d, z :  std_logic_vector(2*K-1 downto 0) := (others=>'0');
  SIGNAL q :  std_logic_vector(K-1 downto 0);
  SIGNAL clk, reset, start, done: std_logic;

  constant DELAY : time := 100 ns;
  constant PERIOD : time := 200 ns;
  constant DUTY_CYCLE : real := 0.5;
  constant OFFSET : time := 0 ns;

BEGIN

  -- Instantiate the Unit Under Test (UUT)
  uut: mult_subt PORT MAP(c => c, d => d, q => q,
                          clk => clk, reset => reset, start => start,
                          z => z, done => done);

  PROCESS -- clock process for clk
  BEGIN
  WAIT for OFFSET;
  CLOCK_LOOP : LOOP
     clk <= '0';
     WAIT FOR (PERIOD *(1.0 - DUTY_CYCLE));
     clk <= '1';
     WAIT FOR (PERIOD * DUTY_CYCLE);
  END LOOP CLOCK_LOOP;
  END PROCESS;

  tb_proc : PROCESS --generate values
    VARIABLE TX_LOC : LINE;
    VARIABLE TX_STR : String(1 to 4096);
    VARIABLE seed1, seed2: positive;
    VARIABLE rand, mul_2k, mul_k: real;
    VARIABLE i_c, i_d, i_q, i_z: integer;

  BEGIN
    mul_2k := real(2**(2*k)); mul_k := real(2**(k)); 
    start <= '0'; reset <= '1';
    WAIT FOR PERIOD;
    reset <= '0';
    WAIT FOR PERIOD;
    for I in 1 to 1000 loop
      UNIFORM(seed1, seed2, rand);
      i_c := INTEGER(TRUNC((0.5-rand) * mul_2k));
      c <= CONV_STD_LOGIC_VECTOR (i_c, 2*k);

      --check that solution c-d*q is in range.
      i_z := 2**(2*K);
      while (i_z > 2**(2*K-1) or i_z < - 2**(2*K-1) ) loop 
        UNIFORM(seed1, seed2, rand);
        i_d := INTEGER(TRUNC((0.5-rand)* mul_2k));
        UNIFORM(seed1, seed2, rand);
        i_q := INTEGER(TRUNC(rand* mul_k));
        i_z := i_c - (i_q * i_d);
      end loop;
      d <= CONV_STD_LOGIC_VECTOR (i_d, 2*k);
      q <= CONV_STD_LOGIC_VECTOR (i_q, k);
      start <= '1';
      WAIT FOR PERIOD;
      start <= '0';
      wait until done = '1';
      WAIT FOR 2*PERIOD;

      IF ( i_z /= (conv_integer(z)) ) THEN 
        write(TX_LOC,string'("ERROR!!! C=")); write(TX_LOC, c);
        write(TX_LOC,string'("- (D=")); write(TX_LOC, d);
        write(TX_LOC,string'("* Q=")); write(TX_LOC, q);
        write(TX_LOC,string'(") is Z=")); write(TX_LOC, z);
        write(TX_LOC,string'(" instead of: ")); write(TX_LOC, (i_z));
        write(TX_LOC, string'("  ( i_c =")); write(TX_LOC, i_c);
        write(TX_LOC, string'(", i_d =")); write(TX_LOC, i_d);
        write(TX_LOC, string'(", i_q =")); write(TX_LOC, i_q);
        write(TX_LOC, string'(" )"));
        TX_STR(TX_LOC.all'range) := TX_LOC.all;
        Deallocate(TX_LOC);
        ASSERT (FALSE) REPORT TX_STR SEVERITY ERROR;
      END IF;  
    end loop;
    WAIT FOR DELAY;
    ASSERT (FALSE) REPORT
    "Simulation successful (not a failure).  No problems detected. "
    SEVERITY FAILURE;
  END PROCESS;

END;
