--------------------------------------------------------------------------------
-- 
-- VHDL Test Bench for module: nr_divider.vhd (ch4) 
--
-- Executes an exhaustive Test Bench. 
--
--------------------------------------------------------------------------------
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE IEEE.std_logic_arith.all;
USE ieee.std_logic_unsigned.all;
USE ieee.numeric_std.ALL;
USE ieee.std_logic_textio.ALL;
USE std.textio.ALL;
use work.nr_divider_parameters.all;

ENTITY test_nr_divider IS
END test_nr_divider;

ARCHITECTURE behavior OF test_nr_divider IS 

  -- Component Declaration for the Unit Under Test (UUT)
  COMPONENT nr_divider is
  PORT (
    a, b: in std_logic_vector(K-1 downto 0);
    clk, reset, start: in std_logic; 
    quotient, remainder: out std_logic_vector(K-1 downto 0);
    done: inout std_logic
    );
  END COMPONENT;

  --Inputs
  SIGNAL a, b :  std_logic_vector(K-1 downto 0) := (others=>'0');
  SIGNAL clk, reset, start, done: std_logic;
  --Outputs
  SIGNAL q, r :  std_logic_vector(K-1 downto 0);

  constant DELAY : time := 100 ns;
  constant PERIOD : time := 200 ns;
  constant DUTY_CYCLE : real := 0.5;
  constant OFFSET : time := 0 ns;

BEGIN

  -- Instantiate the Unit Under Test (UUT)
  uut: nr_divider PORT MAP(a => a, b => b,
                          clk => clk, reset => reset, start => start,
                          quotient => q, remainder => r, done => done);
                          
   PROCESS -- clock process for clk
   BEGIN
    WAIT for OFFSET;
    CLOCK_LOOP : LOOP
       clk <= '0';
       WAIT FOR (PERIOD *(1.0 - DUTY_CYCLE));
       clk <= '1';
       WAIT FOR (PERIOD * DUTY_CYCLE);
    END LOOP CLOCK_LOOP;
   END PROCESS;

  tb_proc : PROCESS --generate values
    VARIABLE TX_LOC : LINE;
    VARIABLE TX_STR : String(1 to 4096);
    BEGIN
      start <= '0'; reset <= '1';
      WAIT FOR PERIOD;
      reset <= '0';
      WAIT FOR PERIOD;
      for J in 1 to 2**K -1 loop
        for I in 0 to 2**K -1 loop
          a <= CONV_STD_LOGIC_VECTOR (I, K);
          b <= CONV_STD_LOGIC_VECTOR (J, K);
          start <= '1';
          WAIT FOR PERIOD;
          start <= '0';
          wait until done = '1';
          WAIT FOR 2*PERIOD;
          IF ( i /= (conv_integer(q)*j + conv_integer(r)) ) THEN 
            write(TX_LOC,string'("ERROR!!! A=")); write(TX_LOC, a);
            write(TX_LOC,string'("/ B=")); write(TX_LOC, b);
            write(TX_LOC,string'(" is Q=")); write(TX_LOC, q);
            write(TX_LOC,string'(" and R=")); write(TX_LOC, r);
            write(TX_LOC,string'(" instead of: ")); write(TX_LOC, (I/J));
            write(TX_LOC,string'(" and: ")); write(TX_LOC, (I mod J));
            write(TX_LOC, string'(" "));
            write(TX_LOC,string'(" (i=")); write(TX_LOC, i);
            write(TX_LOC,string'(" j=")); write(TX_LOC, j); 
            write(TX_LOC, string'(")"));
            TX_STR(TX_LOC.all'range) := TX_LOC.all;
            Deallocate(TX_LOC);
            ASSERT (FALSE) REPORT TX_STR SEVERITY ERROR;
          END IF;  
          end loop;
      end loop;
    WAIT FOR DELAY;
    ASSERT (FALSE) REPORT
    "Simulation successful (not a failure).  No problems detected. "
    SEVERITY FAILURE;
   END PROCESS;

END;
